/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2019, CFWare, LLC
 *
 * Corey Farrell <git@cfware.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Detect the special information tone.
 *
 * \author Corey Farrell <git@cfware.com>
 *
 * \ingroup functions
 */

/*** MODULEINFO
	<support_level>extended</support_level>
 ***/

#include "asterisk.h"
#include "asterisk/channel.h"
#include "asterisk/dsp.h"
#include "asterisk/framehook.h"
#include "asterisk/module.h"
#include "asterisk/pbx.h"

static void datastore_destroy_cb(void *data) {
	ast_free(data);
}

static const struct ast_datastore_info sit_datastore = {
	.type = AST_MODULE,
	.destroy = datastore_destroy_cb
};

static void sit_remove_existing(struct ast_channel *chan)
{
	int *id;
	struct ast_datastore *datastore = ast_channel_datastore_find(chan, &sit_datastore, NULL);

	if (datastore) {
		id = datastore->data;
		ast_framehook_detach(chan, *id);
		ast_channel_datastore_remove(chan, datastore);
		ast_datastore_free(datastore);
	}
}

static void hook_destroy_cb(void *framedata)
{
	ast_dsp_free(framedata);
}

static struct ast_frame *hook_event_cb(struct ast_channel *chan, struct ast_frame *frame, enum ast_framehook_event event, void *data)
{
	struct ast_dsp *dsp = data;

	if (!frame || event != AST_FRAMEHOOK_EVENT_READ) {
		return frame;
	}

	if (frame->frametype == AST_FRAME_CONTROL && frame->subclass.integer == AST_CONTROL_ANSWER) {
		sit_remove_existing(chan);

		return frame;
	}

	return ast_dsp_process(chan, dsp, frame);
}

static int sit_watch_helper(struct ast_channel *chan, const char *cmd, char *data, const char *value)
{
	struct ast_framehook_interface interface = {
		.version = AST_FRAMEHOOK_INTERFACE_VERSION,
		.event_cb = hook_event_cb,
		.destroy_cb = hook_destroy_cb,
	};
	int i;
	struct ast_dsp *dsp;

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}

	dsp = ast_dsp_new();
	if (!dsp) {
		return -1;
	}

	ast_dsp_set_features(dsp, DSP_PROGRESS_BUSY | DSP_PROGRESS_CONGESTION);
	interface.data = dsp;

	ast_channel_lock(chan);
	i = ast_framehook_attach(chan, &interface);
	if (i != -1) {
		int *id;
		struct ast_datastore *datastore;

		sit_remove_existing(chan);

		datastore = ast_datastore_alloc(&sit_datastore, NULL);
		if (!datastore) {
			ast_framehook_detach(chan, i);
			ast_channel_unlock(chan);

			return 0;
		}

		id = ast_calloc(1, sizeof(int));
		if (!id) {
			ast_datastore_free(datastore);
			ast_framehook_detach(chan, i);
			ast_channel_unlock(chan);
			return 0;
		}

		*id = i;
		datastore->data = id;
		ast_channel_datastore_add(chan, datastore);
	}
	ast_channel_unlock(chan);

	return 0;
}

static struct ast_custom_function sit_watch_fn = {
	.name = "SIT_WATCH",
	.write = sit_watch_helper,
};

static int unload_module(void)
{
	return ast_custom_function_unregister(&sit_watch_fn);
}

static int load_module(void)
{
	int res = ast_custom_function_register(&sit_watch_fn);
	return res ? AST_MODULE_LOAD_DECLINE : AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO_STANDARD_EXTENDED(ASTERISK_GPL_KEY, "Detect the special information tone.");
